﻿// 競技者クライアントのサンプルプログラムソースコーC#版 
// このプログラムがいつも意図した動作をするとは限りませんので、自己責任で利用してください。
// Visual C# 2008 Express Editionで対象フレームワークを.NET Framework 2.0に設定して動作確認しました。
// プロキシの設定にはシステムの設定が適用されます。
using System;
using System.Xml;
using System.Threading;

/// <summary>
/// サンプルの競技クライアントクラスです。
/// サーバーによって試合情報(マップなど)が公開された後に実行してください。競技の最中に起動しても問題ありません。
/// 最初に呼ばれるMain()メソッド内のSampleClient作成時に、コンテストIDやチームID、チームトークンを指定してください。
/// もしも競技クライアントの開発にこのクラスを有効活用したければ、instruct()メソッドを確認してください。
/// このメソッドが命令の決定にかかわる部分です。
/// instruct()メソッドを書き換える、またはオーバーライドすることによって命令決定のロジックを組み換えることができます。
/// </summary>
public class SampleClient{
	protected String serverHost;
	protected int contestId;
	protected int teamId;
	protected String teamToken;
	
	/// <summary>試合情報(ゲーム開始時に一度だけ取得する情報)を格納します。</summary>
	protected GameInfo gameInfo;
	/// <summary>試合の状況(セッションごとに命令送信の結果として取得する情報)を格納します。</summary>
	protected BattleState battleState;
	/// <summary>gameInfo内のteamInfo配列や、battleState内のteamState配列の自分のチームに対応するインデックスです。</summary>
	protected int myTeamIndex;
	
	/// <summary>
	/// SampleClientを作成します。 
	/// serverHostにサーバーのホスト名、contestIdにコンテストID、teamIdにチームID、teamTokenにチームトークンを指定してください。
	/// </summary>
	public SampleClient(String serverHost, int contestId, int teamId, String teamToken){
		this.serverHost = serverHost;
		this.contestId = contestId;
		this.teamId = teamId;
		this.teamToken = teamToken;
		gameInfo = new GameInfo();
		battleState = new BattleState();
	}
	
	/// <summary>
	/// 試合情報をダウンロードして、gameInfoに格納します。
	/// ただし、gameInfo内のteamInfoはidの昇順にソートされます。
	/// myTeamIndexにはteamInfoの配列の自分のチームにあたるインデックスが設定されます。
	/// また、battleStateにもチームIDや初期位置が格納されます。
	/// </summary>
	protected virtual bool downloadGameInfo(){
		System.Console.WriteLine("Downloading... ");

		XmlDocument doc = new XmlDocument();
		doc.Load("http://" + serverHost + "/system/map/contest_id=" + contestId);

		XmlNode error = doc.SelectSingleNode("/data/error");
		if(error != null){
			System.Console.WriteLine(error.InnerText);
			return false;
		}
		
		gameInfo.title = doc.SelectSingleNode("/data/title").InnerText;
		gameInfo.totalSession = int.Parse(doc.SelectSingleNode("/data/total_session").InnerText);
		String[] s;
		s = doc.SelectSingleNode("/data/map").InnerText.Split(',');
		for(int i = 0; i < s.Length; i++) gameInfo.map[i] = int.Parse(s[i]);
		XmlNodeList nodes = doc.SelectNodes("/data/site");
		for(int i = 0; i < nodes.Count; i++){
			gameInfo.teamInfo[i] = new TeamInfo();
			gameInfo.teamInfo[i].domain = nodes.Item(i).SelectSingleNode("domain").InnerText;
			gameInfo.teamInfo[i].id = int.Parse(nodes.Item(i).SelectSingleNode("id").InnerText);
			gameInfo.teamInfo[i].name = nodes.Item(i).SelectSingleNode("name").InnerText;
			s = nodes.Item(i).SelectSingleNode("start_position").InnerText.Split(',');
			for(int j = 0; j < s.Length; j++){
				gameInfo.teamInfo[i].startPosition[j] = int.Parse(s[j]);
			}
		}
		Array.Sort(gameInfo.teamInfo);
		for(int i = 0; i < gameInfo.teamInfo.Length; i++){
			if(gameInfo.teamInfo[i].id == teamId) myTeamIndex = i;
		}
		
		for(int i = 0; i < battleState.teamState.Length; i++){
			battleState.teamState[i] = new TeamState();
			battleState.teamState[i].teamId = gameInfo.teamInfo[i].id;
			for(int j = 0; j < 3; j++){
				battleState.teamState[i].machineState[j] = new MachineState();
				battleState.teamState[i].machineState[j].number = j + 1;
				battleState.teamState[i].machineState[j].position = gameInfo.teamInfo[i].startPosition[j];
			}
		}
		return true;
	}
	
	/// <summary>
	/// ロボット3体分の命令を送信して、その結果をbattleStateに格納します。
	/// ただし、battleState内のteamInfoはteam_idの昇順にソートされます。
	/// また、teamInfo内のmachineInfoはそれぞれnumberの昇順にソートされます。
	/// </summary>
	protected virtual void sendInstructions(String[] instructions){
		System.Console.WriteLine("Sending... (" + instructions[0] + "," + instructions[1] + "," + instructions[2] + ")");

		XmlDocument doc = new XmlDocument();
		doc.Load("http://" + serverHost + "/system/command/machine_1=" + instructions[0] + 
			"&machine_2=" + instructions[1] + "&machine_3=" + instructions[2] + 
			"&token=" + teamToken + "&contest_id=" + contestId);

		XmlNode error = doc.SelectSingleNode("/data/error");
		if(error != null) {
			System.Console.WriteLine(error.InnerText);
			return;
		}
		
		battleState = new BattleState();
		battleState.session = int.Parse(doc.SelectSingleNode("/data/session").InnerText);
		String[] s;
		s = doc.SelectSingleNode("/data/map/owner").InnerText.Split(',');
		for(int i = 0; i < s.Length; i++) battleState.owner[i] = int.Parse(s[i]);
		s = doc.SelectSingleNode("/data/map/provision").InnerText.Split(',');
		for(int i = 0; i < s.Length; i++) battleState.provision[i] = int.Parse(s[i]);
		XmlNodeList nodes = doc.SelectNodes("/data/team");
		for(int i = 0; i < nodes.Count; i++) {
			battleState.teamState[i] = new TeamState();
			battleState.teamState[i].teamId = int.Parse(nodes.Item(i).SelectSingleNode("team_id").InnerText);
			XmlNodeList mnodes = nodes.Item(i).SelectNodes("machine");
			for(int j = 0; j < mnodes.Count; j++){
				battleState.teamState[i].machineState[j] = new MachineState();
				battleState.teamState[i].machineState[j].number = int.Parse(mnodes.Item(j).SelectSingleNode("number").InnerText);
				battleState.teamState[i].machineState[j].carryover = int.Parse(mnodes.Item(j).SelectSingleNode("carryover").InnerText);
				battleState.teamState[i].machineState[j].position = int.Parse(mnodes.Item(j).SelectSingleNode("position").InnerText);
				battleState.teamState[i].machineState[j].ryoma = int.Parse(mnodes.Item(j).SelectSingleNode("ryoma").InnerText);
			}
			Array.Sort(battleState.teamState[i].machineState);
		}
		Array.Sort(battleState.teamState);
	}
		
	/// <summary>
	/// ロボット3体分の命令を決定して送信します。セッションごとに呼び出されます。
	/// このメソッドを書き換える、またはオーバーライドすることによって命令決定のロジックを組み換えることができます。
	/// gameInfoとbattleStateを参照しつつ次の命令を考え、sendInstructions()を呼び出すという流れになります。
	/// </summary>
	protected virtual void instruct(){
		// 例として、ロボット1に"712173"、ロボット2に"237422"、ロボット3に"837333"を送信しています。
		String[] instructions = new String[3];
		instructions[0] = "712173";
		instructions[1] = "237422";
		instructions[2] = "837333";
		
		// 例えば次のプログラムを挿入すると、
		// ロボットの左上に敵チームによって配水されたセルがあった時には、左上に移動する動作を行うようになります。
		// for(int i = 0; i < 3; i++){
		// 	int pos = battleState.teamState[myTeamIndex].machineState[i].position; // 現在位置
		// 	int newPos = pos - 42; // 現在位置の左上の位置
		// 	if(newPos >= 0 && battleState.provision[newPos] > 0 && battleState.owner[newPos] != teamId){
		// 		instructions[i] = "6666666";
		// 	}
		// }
		
		// 命令を送信 
		sendInstructions(instructions);
	}
	
	
	/// <summary>
	/// クライアントを実行します。
	/// まず始めに、downloadGameInfo()を呼び出すことで、試合情報をダウンロードします。
	/// その後、繰り返しinstruct()を呼び出すことにより、各セッションの命令を決定して送信します。
	/// 試合中は命令送信のタイミングがセッション更新(3秒間隔)に同期します。
	/// </summary>
	public virtual void run(){
		// 試合情報をダウンロード 
		if(!downloadGameInfo()) return;
		// ダウンロードした試合名とセッション数を表示 
		System.Console.WriteLine("Game Title: " + gameInfo.title);
		System.Console.WriteLine("Total Session: " + gameInfo.totalSession);
		
		while(true){
			// 現在の(処理の終わった)セッション番号を表示 
			System.Console.WriteLine("Processed Session: " + battleState.session);
			if(battleState.session >= gameInfo.totalSession) break;
			// 命令を決定して送信 
			instruct();
			// 1秒待つ 
			Thread.Sleep(1000);
		}
	}
	
	/// <summary>
	/// プログラム実行時、最初に呼ばれるメソッドです。 
	/// SampleClientのインスタンスを作成し、run()を呼び出すことによりクライアントを実行します。
	/// </summary>
    /// 
	public static void Main(String[] args){
		// このSampleClient作成時の引数にサーバーのホスト名、コンテストID、チームID、チームトークンを指定してください。
		new SampleClient("procon-online.net", 1, 1, "abcde").run();
	}
	
	
	
	// --------------------------------------------------------------------------------------------------
	// これ以降は、データ格納用のクラスの定義です。主に、
    // ・試合情報(ゲーム開始時に一度だけ取得するマップなどに関する情報)と、
	// ・試合の状況(セッションごとに命令送信の結果として取得する情報)
	// があります。
	// --------------------------------------------------------------------------------------------------
	
	
	/// <summary>
	/// 試合情報を格納するクラスです。 
	/// </summary>
	public class GameInfo{
		public String title;
		public int totalSession;
		public int[] map = new int[41 * 41];
		public TeamInfo[] teamInfo = new TeamInfo[6];
	}

	/// <summary>
	/// 試合情報(GameInfo)の中の、チームごとの情報を格納するクラスです。 
	/// </summary>
	public class TeamInfo :IComparable<TeamInfo>{
		public String domain;
		public int id;
		public String name;
		public int[] startPosition = new int[3];
		
		public int CompareTo(TeamInfo a){
			return id - a.id;
		}
	}

	/// <summary>
	/// 試合の状況を格納するクラスです。 
	/// </summary>
	public class BattleState{
		public int session;
		public int[] owner = new int[41 * 41];
		public int[] provision = new int[41 * 41];
		public TeamState[] teamState = new TeamState[6];
	}

	/// <summary>
	/// 試合の状況(BattleState)の中で、チームごとの状況を格納するクラスです。 
	/// </summary>
	public class TeamState :IComparable<TeamState>{
		public int teamId;
		public MachineState[] machineState = new MachineState[3];
		
		public int CompareTo(TeamState a){
			return teamId - a.teamId;
		}
	}

	/// <summary>
	/// 試合の状況(BattleState)の中の、チームごとの状況(TeamState)の中で、ロボットごとの状況を格納するクラスです。 
	/// </summary>
	public class MachineState :IComparable<MachineState>{
		public int carryover;
		public int number;
		public int position;
		public int ryoma;
		
		public int CompareTo(MachineState a){
			return number - a.number;
		}
	}
}
